/* $Id$ */

#ifndef _JAFAR_IMAGE_IMAGE_HPP_
#define _JAFAR_IMAGE_IMAGE_HPP_

#include <fstream>  
#include <iostream>  

#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include "boost/archive/xml_iarchive.hpp"
#include "boost/archive/xml_oarchive.hpp"
#include "boost/archive/binary_iarchive.hpp"
#include "boost/archive/binary_oarchive.hpp"
#include "boost/serialization/split_member.hpp"
#include "boost/serialization/binary_object.hpp"
#include <boost/serialization/string.hpp>
#include "boost/shared_ptr.hpp"

#include <kernel/jafarMacro.hpp>
#include "image/cImage.h"
#include <opencv/cxcore.h>
#if CV_MAJOR_VERSION*10 + CV_MINOR_VERSION > 21
#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/legacy/legacy.hpp>
#else
#include <cxcore.hpp> // opencv
#include <cvaux.hpp>
#include <highgui.hpp>
#endif

#ifndef SWIG
namespace cv {
std::ostream& operator<<(std::ostream& os, const cv::Rect &r);
std::ostream& operator<<(std::ostream& os, const cv::Point &p);
std::ostream& operator<<(std::ostream& os, const cv::Size &s);
}
#endif

namespace jafar {

  /// namespace for module image
  namespace image {
    void getQuadrangleSubPix(const cv::Mat& src, cv::Mat& dst, const cv::Mat& mapMatrix);

    class Image;
    typedef boost::shared_ptr<Image>  image_ptr_t;
    
    /*! Image class.
     *
     * \ingroup image
     */
    class Image {
      // Cubic kernels generated with generateCubicKernel
      static const int cubicKernelResolution = 255;
      static const double cubicKernel1[256];
      static const double cubicKernel2[256];
      ///OpenCV matrix that holds image data
      cv::Mat m_matrix;
    public:
      Image() { }
      ~Image();
      /**
       * Construct a new Image.
       * @param width the image width
       * @param height the image height
       * @param depth the image depth
       * @param colorSpace the color space of the image
       * @pre \a width_ > 0
       * @pre \a height_ > 0
       * @pre \a depth_ is one supported by OpenCV (CV_8U, CV_8S, CV_16U, CV_16S, CV_32S, CV_32F, CV_64F)
       */
      Image(int width, int height, int depth, JfrImage_TypeColorSpace colorSpace);
      /// copy constructor from Image
      Image( Image const & img );
      /// copy constructor from Image
      Image& operator=( Image const & img );
      /// copy constructor from cv::Mat affects only the data, the header is kept
      Image& operator=( const cv::Mat& matrix );
      /// Override copy constructor of CvImage
      Image( const CvImage& img );
      /// Override constructor from IplImage
      Image( IplImage* img );
      /// Override constructor from cv::Mat
      Image( const cv::Mat& matrix );
      /// copy constructor with relative ROI, without reallocation
      /// @warning if img already has a ROI defined, the two ROIs will be composed
      Image(const Image& img, const cv::Rect& roi);

      /// clone Image (ie copy data)
      Image clone() const;

      
      /// Copies an image into another applying a mask
      void copyTo(image::Image& dst, 
		  const image::Image& mask = image::Image()) const;

      /*! create data (check if allocation or reallocation is needed)
       * @param width the image width
       * @param height the image height
       * @param depth the image depth
       * @param colorSpace the color space of the image
       * @pre \a width_ > 0
       * @pre \a height_ > 0
       * @pre \a depth_ is one supported by OpenCV (CV_8U, CV_8S, CV_16U, CV_16S, CV_32S, CV_32F, CV_64F)
       */
      void create( int width, int height, int depth, JfrImage_TypeColorSpace colorSpace );
      
      /** Allocate and load an image.
       * @param filename is the name of file to be loaded
       * @param iscolor specifies colorness of the loaded image:
       * if >0, the loaded image is forced to be color 3-channel image;
       * if 0, the loaded image is forced to be grayscale;
       * if <0, the loaded image will be loaded as is (with number of channels depends on the file). 
       *
       * Currently the following file formats are supported:
       * Windows bitmaps - BMP, DIB;
       * JPEG files - JPEG, JPG, JPE;
       * JPEG 2000 files - JP2
       * Portable Network Graphics - PNG;
       * Portable image format - PBM, PGM, PPM;
       * Sun rasters - SR, RAS;
       * TIFF files - TIFF, TIF. 
       */
      static Image* loadImage( const char* filename, int iscolor = -1);
      
      /** Load an image.
       * @param filename is the name of file to be loaded
       * @param iscolor specifies colorness of the loaded image:
       * if >0, the loaded image is forced to be color 3-channel image;
       * if 0, the loaded image is forced to be grayscale;
       * if <0, the loaded image will be loaded as is (with number of channels depends on the file). 
       *
       * Currently the following file formats are supported:
       * Windows bitmaps - BMP, DIB;
       * JPEG files - JPEG, JPG, JPE;
       * JPEG 2000 files - JP2
       * Portable Network Graphics - PNG;
       * Portable image format - PBM, PGM, PPM;
       * Sun rasters - SR, RAS;
       * TIFF files - TIFF, TIF. 
       */
      bool load( const std::string& filename, int iscolor = -1 );
      
			/**
				* OpenCV save function
				* @param filename name of the file where to save the image
				*                 (for instance "myimage.png" or "myimage.tiff"...)
			*/
			void save( const std::string& filename) const {
				//std::string name(filename);
				cv::imwrite(filename, m_matrix);
			}
			/**
				* @param imgname it's the name of the image, and is only usefull when saving to
				*                opencv's specific file format (no extension)
				*/
			void save( const std::string& filename, const std::string& imgname ) {
				//std::string name(filename);
				cv::imwrite(filename, m_matrix);
				m_name = imgname;
			}
			/**
				* @param params opencv save parameters
				*/
			void save( const std::string& filename, const std::vector<int>& params ) {
				cv::imwrite(filename, m_matrix, params);
				m_name = filename;
			};

      void print() { JFR_DEBUG(*this); }

      /** Convert Color space of a JfrImage.
       * @param dst_ destination image
       * @param colorCode_ type of conversion.
       * @pre \a dst_ exists
       * \bug Does not update colorSpace attribute.
       */
      void convertColorMode(jafar::image::Image& dst_) const;
      /**
       * Change the depth of the image
       * @pre \a dst_ and this have the same color mode
       */
      void convertDepth(jafar::image::Image& dst_, double scale = 1.0, double offset = 0.0, bool _abs = false) const;
      
      /** Resize the current image in \a resizedImage.
       */
      void resize(Image& resizedImage, int interpolation = CV_INTER_LINEAR) const;
      
      /** Resize the current image in a newly allocated image.
       * @param scale scale factor
       * @param interpolation interpolation mode
       * @return the resized image
       */
      Image* resize(double scale, int interpolation = CV_INTER_LINEAR) const;
      
      /** Resize the current image in a newly allocated image.
       * @param width new width
       * @param height new height
       * @param interpolation interpolation mode
       * @return the resized image
       */
      Image* resize(int newWidth, int newHeight, int interpolation = CV_INTER_LINEAR) const;
      
      /** In-place resize of the current image
       * @param width new width
       * @param height new height
       * @param interpolation interpolation mode 
       */
      void resize(int newWidth, int newHeight, int interpolation = CV_INTER_LINEAR);
      
      /**
       * @return the size of the image data in bytes
       */
      int imageSize() const;
      
      /** @return the colorspace of the image
       */
      inline JfrImage_TypeColorSpace colorSpace() const {
        return header.colorSpace;
      };
      void setColorSpace(JfrImage_TypeColorSpace cs);
      /**
       * @param line the line number
       * @return a pointer to the data of the image
       */
      inline uchar* data(int line =0) {
        return m_matrix.ptr(line);
      };
      
      /**
       * @param line the line number
       * @return a pointer to the data of the image
       */
      inline const uchar* data(int line =0) const {
        return m_matrix.ptr(line);
      };
      
      /**
       * Copy the image in the destination.
       * @param dst destination
       * @param x_src x coordinate in the source
       * @param y_src y coordinate in the source
       * @param x_dst x coordinate in the destination
       * @param y_dst y coordinate in the destination
       */
      void copy( jafar::image::Image & dst, int x_src = 0, int y_src = 0, int x_dst = 0, int y_dst = 0 , int width = -1, int height = -1) const;
      /**
        Same as copy, but deals correctly with parts outside of images (fills black in dst)
        @return true if there isn't any hole in dst (no area filled in black)
      */
      bool robustCopy( jafar::image::Image & dst, int x_src = 0, int y_src = 0, int x_dst = 0, int y_dst = 0 , int width = -1, int height = -1) const;

			bool extractPatch(image::Image &dst, int x, int y, int width, int height) const;
			
      /**
	 Get the absolute Region Of Interest of the image
			
	 @note{General information about ROI since Opencv 2:
	 With opencv 2, ROIs are implicit. It means that once you have set the ROI of an image,
	 the image appears as if it is only the ROI area, so by manipulating what appears 
	 to be the whole image you only manipulate the ROI area.}
			
	 @warning This function is only informative, 
	 it should not be used to take into account the processed area.
      */
      cv::Rect getROI() const
      {
	cv::Size size;
	return getROI(size);
      }
      /**
       * Get the absolute Region Of Interest and full size of the image
       * @param size the full size of the image
       */
      cv::Rect getROI(cv::Size &size) const
      {
	//return cvROIToRect((*this)->roi);
	//return roi();
				
	cv::Point offset;
	m_matrix.locateROI(size, offset);
	return cv::Rect(offset.x,offset.y,m_matrix.size().width,m_matrix.size().height);
      }
      /**
       * Set the absolute Region Of Interest of the image.
       *		
       * Do not forget that you can also create a new image with a relative ROIs
       * without reallocation
       */
      void setROI(const cv::Rect &rect)
      {
	//cvSetImageROI(*this, rect);
	//set_roi(rect);
				
	setROI(rect.x, rect.y, rect.width, rect.height);
      }
      /**
       * Set the absolute Region Of Interest of the image.
       */
      void setROI(int x, int y, int w, int h)
      {
				
	cv::Size size; cv::Point offset;
	m_matrix.locateROI(size, offset);
	m_matrix.adjustROI(offset.y-y, (h+y)-(m_matrix.size().height+offset.y),
			   offset.x-x, (w+x)-(m_matrix.size().width +offset.x));
      }
      /**
       * Reset the Region Of Interest of the image to the full image.
       */
      void resetROI()
      {
	//cvResetImageROI(*this);
	//reset_roi();
				
	cv::Size size; cv::Point offset;
	m_matrix.locateROI(size, offset);
	m_matrix.adjustROI(offset.y,size.height-(m_matrix.size().height+offset.y),
			   offset.x,size.width-(m_matrix.size().width+offset.x));
      }

			
      /**
       * @return the image file name.
       */
      inline std::string name() const {
        return m_name;
      }

      /**
       * @set image file name.
       * @param _name file name of the image
       */
      void setName(const std::string& _name) {
        m_name = _name;
      }

      /** @return the value of a pixel
       * @param channeltype the channeltype of the image
       * @param x_ x-coord of the pixel
       * @param y_ y-coord of the pixel
       * @param channel_ required channel (default 0)
       * @pre \a channeltype must be the real channel type of the image
       * @pre \a x_ and y_ must be inside the image
       */
      template<typename channeltype>
      inline channeltype getPixelValue(int x_, int y_, int channel_=0) const;
      
      /** This function return the subvalue of a pixel with linear
       * interpolation.
       * @return the subvalue of a pixel
       * @param channeltype the channeltype of the image
       * @param x_ x-coord of the subpixel
       * @param y_ y-coord of the subpixel
       * @param channel_ required channel (default 0)
       * @pre \a channeltype must be the real channel type of the image
       * @pre \a x_ and y_ must be inside the image
       * 
       * @warning the coordinates are translated by 0.5, meaning that getSubPixelValue( x_, y_ ) == getPixelValue( x_, y_)
       */
      template<typename channeltype>
      double getSubPixelValue(double x_, double y_, int channel_=0, JfrImage_InterpolationType _interpolation = JfrImage_INTERP_BILINEAR) const;
      
      /** Set the value of a pixel
       * @param channeltype the channeltype of the image
       * @param val_ the new value 
       * @param x_ x-coord of the pixel
       * @param y_ y-coord of the pixel
       * @param channel_ required channel (default 0)
       * @pre \a channeltype must be the real channel type of the image
       * @pre \a x_ and y_ must be inside the image
       */
      template<typename channeltype>
      void setPixelValue(channeltype val_, int x_, int y_, int channel_ = 0);
      
      /**
       * This function compute the ZNCC between two images of the same depth and size
       * @param im2 the second image
       */
      double computeZncc(Image const* im2) const;
      /**
	 This function computes the ZNCC between two images of the same depth in two regions of interest that must have the
	 same size, and can apply a spatially distributed weight. It ignores pixels with extremum value
	 (eg 0 and 255 for char), and ignores parts outside of images, but requires at least half of pixels not ignored
	 to return non 0.
	 @param im2 the second image
	 @param roi1_ the region of interest in this image
	 @param roi2_ the region of interest in the second image (that must have the same size than roi1_)
	 @param weightMatrix the weight matrix that must have the same size than the regions of interest
	 @return correlation score between 0 and 1
      */
      double computeZnccRoisWeight(Image const* im2, cv::Rect const* roi1_ = NULL, cv::Rect const* roi2_ = NULL, float const* weightMatrix = NULL ) const;
      /**
       * This function computes the ZNCC between two images of the same depth in the 
       * given region of interest only.
       * @param im2 the second image
       * @param rectRoi defines the roi
       */
      double computeZnccRoi(Image const* im2, cv::Rect const* rectRoi) const;
      
      /**
       * This function return the best ZNCC between two images, the second images get rotated
       * @param im2 the second image
       * @param rotationStep the step used for the applied rotation
       */
      double computeRotatingZncc( Image const* im2, int rotationStep) const;
      
      /** Apply rank transform in a newly allocated image.
       * @param radius radius of the window used to compute rank
       * @return the filtered image
       */
      Image* rankTransform (const int radius);

      /**
	 This function rotates and scales the image
	 @param angle the rotation angle (degrees)
	 @param factor the scaling factor
	 @param dest the destination image
      */
      void rotateScale(double angle, double factor, Image &dest) const;

      /**
       * @return the orientation of the image, computed from the gradients.
       */
      double orientation(double sigma = 1.5);
      
      /*!
       * @return true if img has the same characteristics than this
       */
      bool compHeader(Image const & img) const {
	return (width() == img.width() && 
		height() == img.height() &&
		depth() == img.depth() &&
		colorSpace() == img.colorSpace());
      }
      ///convert to cv::Mat reference
      operator cv::Mat& () { return this->m_matrix; }
      ///convert to const cv::Mat reference
      operator const cv::Mat& () const { return this->m_matrix; }
      ///convert to const IplImage
      operator const IplImage () const { 
        return (const IplImage)m_matrix;
      }
      ///convert to IplImage
      operator IplImage () { 
	return (IplImage)m_matrix;
      }
      ///convert to const CvMat
      operator const CvMat () const { return (const CvMat)m_matrix; }
      ///convert to CvMat
      operator CvMat () { return (CvMat)m_matrix; }
      ///@return the image's matrix size
      inline cv::Size size() const { return m_matrix.size(); }
      ///@return whether the image is empty or not
      inline bool empty() const { return m_matrix.empty(); }
      ///@return the image's width
      inline int width() const { return m_matrix.cols; }
      ///@return the image's height
      inline int height() const { return m_matrix.rows; }
      ///@return the image's step in bytes
      inline size_t step() const { return m_matrix.step; }
      ///@return the image's step in number of elements
      inline size_t step1() const { return m_matrix.step1(); }
      ///@return the image's depth
      inline int depth() const { return m_matrix.depth(); }
      ///@return the image's channels number
      inline int channels() const { return m_matrix.channels(); }
      ///@return the image's matrix type
      inline int type() const { return m_matrix.type(); }
      ///@return the image's data organized in cv::Mat
      inline cv::Mat mat() const { return m_matrix; }
      ///detaches image
      inline void detach() { m_matrix.release(); }
      ///pixel size
      inline size_t pix_size() { return !m_matrix.empty() ? m_matrix.elemSize() : 0; }
      ///set the matrix elements to the given value
      void setTo(const cv::Scalar& s, const image::Image& mask=image::Image()) {
	m_matrix.setTo(s, mask);
      }
      ///convert image to another one, it is a wrapping to cv::Mat::converTo
      void convertTo(image::Image& dst, double alpha=1, double beta=0) const {
	m_matrix.convertTo(dst, dst.type(), alpha, beta);
      }

    private:
      
      void initImageHeader();
      // serialization
      friend class boost::serialization::access;
			
      template<class Archive>
      void save(Archive & ar, const unsigned int version) const {
	JFR_PRECOND(!m_matrix.empty(), "Nothing to save");
	IplImage image = (IplImage)m_matrix;
	ar << boost::serialization::make_nvp("width",image.width);  
	ar << boost::serialization::make_nvp("height",image.height);  
	ar << boost::serialization::make_nvp("depth",image.depth);
	ar << boost::serialization::make_nvp("colorSpace",header.colorSpace); 
	ar << boost::serialization::make_nvp("nChannels",image.nChannels);
	ar << boost::serialization::make_nvp("nSize",image.nSize);  
	ar << boost::serialization::make_nvp("ID",image.ID);   
	ar << boost::serialization::make_nvp("dataOrder",image.dataOrder);  
	ar << boost::serialization::make_nvp("origin",image.origin);   
	ar << boost::serialization::make_nvp("imageSize",image.imageSize);  
	ar << boost::serialization::make_nvp("pixeldata", boost::serialization::make_binary_object(image.imageData, image.imageSize));
	ar << boost::serialization::make_nvp("widthStep",image.widthStep);
      }
      
      template<class Archive>
      void load(Archive & ar, const unsigned int version) {
	int width, height, depth;
	JfrImage_TypeColorSpace cs;
	ar >> boost::serialization::make_nvp("width",width);
	ar >> boost::serialization::make_nvp("height",height);
	ar >> boost::serialization::make_nvp("depth",depth);
	ar >> boost::serialization::make_nvp("colorSpace",cs);
	create(width, height, depth, cs);
	IplImage image = (IplImage)m_matrix;
	ar >> boost::serialization::make_nvp("nChannels",image.nChannels);
	ar >> boost::serialization::make_nvp("nSize",image.nSize);  
	ar >> boost::serialization::make_nvp("ID",image.ID);
	ar >> boost::serialization::make_nvp("dataOrder",image.dataOrder);  
	ar >> boost::serialization::make_nvp("origin",image.origin);
	ar >> boost::serialization::make_nvp("imageSize",image.imageSize);
	ar >> boost::serialization::make_nvp("pixeldata", boost::serialization::make_binary_object(image.imageData, image.imageSize));  
	ar >> boost::serialization::make_nvp("widthStep",image.widthStep);  
	memset(image.BorderMode, 0, 4);  
	memset(image.BorderConst, 0, 4);  
	image.imageDataOrigin = image.imageData;
      }
#ifndef SWIG
      BOOST_SERIALIZATION_SPLIT_MEMBER()
#endif /* NOT SWIG */

    private:
      /**
       * Call as \p t - \p dt
       * @param t
       * @param dt 
       */
      inline double cubicKernel( int t, double dt ) const;
      inline double cubicKernelI( int t, int idx ) const;
    private:
      JfrImage_ImageHeader header;

      std::string m_name;
    }; // class Image
    
    template<typename channeltype>
    channeltype Image::getPixelValue(int x_, int y_, int channel_) const {
      JFR_PRECOND(x_ >= 0 && y_ >= 0 && x_ < width() && y_ < height(), "JfrImage::pixelValue pixel coordinates are not in the image.");
      JFR_PRECOND( 0 <= channel_ && channel_ < channels(), "JfrImage::pixelValue wrong channel parameter ");
      channeltype const *pxPtr;
      pxPtr = reinterpret_cast<channeltype const *>(data(y_));
      return *(pxPtr + x_*channels() + channel_);
    }

    inline double Image::cubicKernelI( int _t, int _idx ) const
    {
      switch( _t )
      {
        case -1:
          return cubicKernel2[_idx];
        case 0:
          return cubicKernel1[_idx];
        case 1:
          return cubicKernel1[cubicKernelResolution - _idx];
        case 2:
          return cubicKernel2[cubicKernelResolution - _idx];
      }
      return 0;
    }
    inline double Image::cubicKernel( int _t, double _dt ) const
    {
      int idx =int (_dt * cubicKernelResolution );
      return cubicKernelI( _t, idx );
    }
    
    template<typename channeltype>
    double Image::getSubPixelValue(double x_, double y_, int channel_, JfrImage_InterpolationType _interpolation ) const {
      int height_ = height();
      int width_ = width();
      JFR_PRECOND(x_ >= 0. && y_ >= 0. && x_ <= (double)width_ - 1 && y_ <= (double)height_ - 1, 
                  "JfrImage::pixelValue pixel coordinates are not in the image.");
      JFR_PRECOND( 0 <= channel_ && channel_ < channels(), "JfrImage::getSubpixelValue wrong channel parameter ");
      if( x_ == width() - 1) x_ -= 1e-6;
      if( y_ == height() - 1) y_ -= 1e-6;
  
      switch( _interpolation )
      {
        case JfrImage_INTERP_NEAREST:
        {
          return getPixelValue<channeltype>( (int)nearbyint( x_ ), (int)nearbyint( y_ ), channel_ );
        }
        case JfrImage_INTERP_BILINEAR:
        {
          int xi = (int)floor(x_);
          int yi = (int)floor(y_);
      
          double dx = x_-floor(x_);
          double dy = y_-floor(y_);
          double c11 = (1-dx)*(1-dy);
          double c12 = (1-dy)*dx;
          double c21 = (1-dx)*dy;
          double c22 = dx*dy;
      
          return (
              c11 * getPixelValue<channeltype>(xi,yi, channel_)
              + c12 * getPixelValue<channeltype>(xi+1,yi, channel_)
              + c21 * getPixelValue<channeltype>(xi,yi+1, channel_)
              + c22 * getPixelValue<channeltype>(xi+1,yi+1, channel_));
        }
        case JfrImage_INTERP_CUBIC:
        {
          double result = 0.0;
          int xi = int(x_);
          int yi = int(y_);
          double dx = x_ - xi;
          int idxX =int( dx * cubicKernelResolution );
          double dy = y_ - yi;
          int idxY =int( dy * cubicKernelResolution );
          int channelsCount = channels();
          int offset = (xi - 1) * channelsCount + channel_;
          for(int v = -1; v < 3; ++v)
          {
            int yt = yi + v;
            if( yt >= 0 and yt < height_ )
            {
              channeltype const *pxPtr = reinterpret_cast<channeltype const *>(data(yt));
              pxPtr += offset;
              
              double ycoef = cubicKernelI( v, idxY);
              for(int u = -1; u < 3; ++u, pxPtr += channelsCount)
              {
                int xt = xi + u;
                if( xt >= 0 and xt < width_ ) // TODO use a better value than 0 when out of range
                {
                  result += ycoef * cubicKernelI( u, idxX ) * *pxPtr;
                }
              }
            }
          }
          return result;
        }
      }
      return -1;
    }
    
    template<typename channeltype>
    void Image::setPixelValue(channeltype val_, int x_, int y_, int channel_) {
      JFR_PRECOND(x_ >= 0 && y_ >= 0 && x_ < width() && y_ < height(), 
          "JfrImage::pixelValue pixel coordinates are not in the image.");
      JFR_PRECOND( 0 <= channel_ && channel_ < channels(), "JfrImage::pixelValue "
          "wrong channel parameter ");
      channeltype *pxPtr = reinterpret_cast<channeltype*>(data(y_));
      *(pxPtr+ x_ * channels() + channel_) = val_;
    }

    std::ostream& operator<<(std::ostream& s, const Image& image);
    std::ostream& operator<<(std::ostream& s, const JfrImage_TypeColorSpace tcs);

    /** I/O methods for Image serialization to file.
     * strange behavior of boost serialization with functions in namespace, due to templating.
     * so we use a class.
     * moreover, using those functions directly in the class Image broke the serialization when Image 
     * is a member of another serializable class. (surely I don't understand how to do this properly)
     *
     * WARNING: boost documentation announce non portability for binary archive for the current release (1.33.1)
     * \ingroup image
     */
    class ImageIO {
    public:
      static void xml_load(jafar::image::Image & img, std::string const & filename);
      static void xml_save(jafar::image::Image & img, std::string const & filename);
      static void bin_load(jafar::image::Image & img, std::string const & filename);
      static void bin_save(jafar::image::Image & img, std::string const & filename);
    private:
      ImageIO() {};
    }; /* ImageIO */
    
  } /* image */
} /* jafar */

#endif
